#include <gmp.h>
#include <stdio.h>
#include <string.h>

#define mpz_dec(x) mpz_sub_ui(x, x, 1);
#define mpz_inc(x) mpz_add_ui(x, x, 1);

int main (int argc, char *argv[])
{
	mpz_t n, e, p, q, pq, d, m;
	
	// Initialize all of the variables and set the ones we can.
	mpz_init_set_str(n, "c3b3a7015c04299ff3a25f104e2285c1ec2d55471e6208959d0f6981b2fa2c6d3e316f9364d5eb5c7789e142b75bfaf402e7e02fac0cb09f6419db1f44679f8bbcca142f1d312feb095708ef175a4ef80271321e7240f0d854c90a74fc59209cdf80aa8f85ae3b948a3ce55c69cd050098d5a79aebbc241cc642b106b1af2cb7", 16);
	mpz_init_set_si(e, 65537);
	mpz_init(p);
	mpz_init(q);
	mpz_init(pq);
	mpz_init(d);
	mpz_init(m);
	
	// Take the square root of 'n' and place the result in 'p'.
	mpz_sqrt(p, n);
	
	// Take the result of 'n mod p' and store it in 'm' for an initial test.
	mpz_mod(m, n, p);
	
	while(mpz_cmp_si(m, 0))
	{
		// Loop through prime numbers until 'n mod p' is equal to zero.
		mpz_nextprime(p, p);
		gmp_printf(": p=%#Zx\r", p);
		mpz_mod(m, n, p);
	}
	
	// 'q' was found, let's print the good news.
	printf("\n::: Found 'p' and 'q'!!!\n");
	printf(":: Calculating private key...\n");
	
	// Calculate the private key.
	mpz_tdiv_q(q, n, p);	// Divide 'n' by 'p' and store the result in 'q'. (The division truncates)
	mpz_sub_ui(pq, p, 1);	// Subtruct one from 'p' and store the result in 'pq'.
	mpz_dec(q);		// Decrement 'q' by one.
	mpz_mul(pq, pq, q);	// Get the result of '(q-1)(p-1)'.
	mpz_inc(q);		// Increment 'q'.
	mpz_invert(d, e, pq);	// Get the mod inverse of 'e' and 'pq' and store it in 'd'. (This get's the private key).
	
	// Finally, output all the values.
	gmp_printf(": n = %#Zx\n", n);
	gmp_printf(": p = %#Zx\n", p);
	gmp_printf(": q = %#Zx\n", q);
	gmp_printf(": e = %#Zx\n", e);
	gmp_printf(": d = %#Zx\n", d);
	
	printf(":: Saving to file 'key.txt'...\n");
	
	FILE *fp;
	fp = fopen("key.txt", "w");
	
	// Check that we were able to open a file stream.
	if(fp != NULL)
	{
		// If we were, start saving the data
		fprintf(fp, "n = ");
		mpz_out_str(fp, 16, n);
		
		fprintf(fp, "\np = ");
		mpz_out_str(fp, 16, p);
		
		fprintf(fp, "\nq = ");
		mpz_out_str(fp, 16, q);
		
		fprintf(fp, "\ne = ");
		mpz_out_str(fp, 16, e);
		
		fprintf(fp, "\nd = ");
		mpz_out_str(fp, 16, d);
		
		fclose(fp);
	}
	else
	{
		// Otherwise, alert the user that the program will not quit and
		// will stay open until they close it.
		
		printf("::: ERROR: Unable to save to file!!!\n");
		printf(":: This program will stay open until explicitly closed.");

		// Yes, I know I could have done while(1); but I just felt like putting this in here.
		__asm("loop:\nnop\njmp loop");
	}
	
	return 0;
}

